from .routing_service import RoutingService
from .zmq_proxy_router import serialize_frames, deserialize_frames
from .external_rpc_service import ExternalRPCService
from .pubsub_service import PubSubService
from .zmq_proxy_router import ZMQProxyRouter



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import logging as _log

import gevent

from .core import *
from .errors import *
from .decorators import *
from .subsystems import *

from volttron.utils import is_valid_identity, get_address, ClientContext as cc
from volttron.utils.keystore import get_server_keys


class Agent(object):

    class Subsystems(object):

        def __init__(
            self,
            owner,
            core,
            heartbeat_autostart,
            heartbeat_period,
            enable_store,
            enable_web,
            enable_channel,
            message_bus,
        ):
            self.peerlist = PeerList(core)
            self.ping = Ping(core)
            self.rpc = RPC(core, owner, self.peerlist)
            self.hello = Hello(core)
            if message_bus == "rmq":
                self.pubsub = RMQPubSub(core, self.rpc, self.peerlist, owner)
            else:
                self.pubsub = PubSub(core, self.rpc, self.peerlist, owner)
                # Available only for ZMQ agents
                if enable_channel:
                    self.channel = Channel(core)
            self.health = Health(owner, core, self.rpc)
            self.heartbeat = Heartbeat(
                owner,
                core,
                self.rpc,
                self.pubsub,
                heartbeat_autostart,
                heartbeat_period,
            )
            if enable_store:
                self.config = ConfigStore(owner, core, self.rpc)
            if enable_web:
                self.web = WebSubSystem(owner, core, self.rpc)
            self.auth = Auth(owner, core, self.rpc)

    def __init__(
        self,
        identity=None,
        address=None,
        context=None,
        publickey=None,
        secretkey=None,
        serverkey=None,
    # Since heartbeat is now 100% tied to status on the vctl change the defaults
    # to auto start the heartbeat.
        heartbeat_autostart=True,
        heartbeat_period=60,
        volttron_home=None,
        agent_uuid=None,
        enable_store=True,
        enable_web=False,
        enable_channel=False,
        reconnect_interval=None,
        version="0.1",
        instance_name=None,
        message_bus=None,
        volttron_central_address=None,
        volttron_central_instance_name=None,
    ):

        if volttron_home is None:
            volttron_home = cc.get_volttron_home()

        try:
            self._version = version

            if identity is not None and not is_valid_identity(identity):
                _log.warning("Deprecation warning")
                _log.warning("All characters in {identity} are not in the valid set.".format(
                    identity=identity))

            if message_bus is not None and message_bus.lower() == "rmq":
                _log.debug("Creating RMQ Core {}".format(identity))
                self.core = RMQCore(
                    self,
                    identity=identity,
                    address=address,
                    context=context,
                    publickey=publickey,
                    secretkey=secretkey,
                    serverkey=serverkey,
                    instance_name=instance_name,
                    volttron_home=volttron_home,
                    agent_uuid=agent_uuid,
                    reconnect_interval=reconnect_interval,
                    version=version,
                    volttron_central_address=volttron_central_address,
                    volttron_central_instance_name=volttron_central_instance_name,
                )
            else:
                _log.debug("Creating ZMQ Core {}".format(identity))
                self.core = ZMQCore(
                    self,
                    identity=identity,
                    address=address,
                    context=context,
                    publickey=publickey,
                    secretkey=secretkey,
                    serverkey=serverkey,
                    instance_name=instance_name,
                    volttron_home=volttron_home,
                    agent_uuid=agent_uuid,
                    reconnect_interval=reconnect_interval,
                    version=version,
                )
            self.vip = Agent.Subsystems(
                self,
                self.core,
                heartbeat_autostart,
                heartbeat_period,
                enable_store,
                enable_web,
                enable_channel,
                message_bus,
            )
            self.core.setup()
            self.vip.rpc.export(self.core.version, "agent.version")
        except Exception as e:
            _log.exception("Exception creating Agent. {}".format(e))
            raise e


class BasicAgent(object):

    def __init__(self, **kwargs):
        kwargs.pop("identity", None)
        super(BasicAgent, self).__init__(**kwargs)
        self.core = BasicCore(self)


def build_agent(address=None,
                identity=None,
                publickey=None,
                secretkey=None,
                timeout=10,
                serverkey=None,
                agent_class=Agent,
                volttron_central_address=None,
                volttron_central_instance_name=None,
                **kwargs) -> Agent:
    """Builds a dynamic agent connected to the specifiedd address.

    All key parameters should have been encoded with
    :py:meth:`volttron.client.vip.socket.encode_key`

    :param str address: VIP address to connect to
    :param str identity: Agent's identity
    :param str publickey: Agent's Base64-encoded CURVE public key
    :param str secretkey: Agent's Base64-encoded CURVE secret key
    :param str serverkey: Server's Base64-encoded CURVE public key
    :param class agent_class: Class to use for creating the instance
    :param int timeout: Seconds to wait for agent to start
    :param kwargs: Any Agent specific parameters
    :return: an agent based upon agent_class that has been started
    :rtype: agent_class
    """

    address = address if address is not None else get_address()

    # This is a fix allows the connect to message bus to be different than
    # the one that is currently running.
    if publickey is None or secretkey is None:
        publickey, secretkey = get_server_keys()

    message_bus = cc.get_messagebus()

    try:
        enable_store = kwargs.pop("enable_store")
    except KeyError:
        enable_store = False

    agent = agent_class(address=address,
                        identity=identity,
                        publickey=publickey,
                        secretkey=secretkey,
                        serverkey=serverkey,
                        volttron_central_address=volttron_central_address,
                        volttron_central_instance_name=volttron_central_instance_name,
                        message_bus=message_bus,
                        enable_store=enable_store,
                        **kwargs)
    event = gevent.event.Event()
    gevent.spawn(agent.core.run, event)
    with gevent.Timeout(timeout):
        event.wait()
    return agent



# Module copied from
# http://code.activestate.com/recipes/576642-persistent-dict-with-multiple-standard-file-format/
import csv
import os
import shutil
import logging
import pickle
from threading import Thread
from queue import Queue
from copy import deepcopy

from volttron.utils import jsonapi

_log = logging.getLogger(__name__)


def load_create_store(filename):
    persist = PersistentDict(filename=filename, flag="c", format="json")
    return persist


class PersistentDict(dict):
    """Persistent dictionary with an API compatible with shelve and anydbm.

    The dict is kept in memory, so the dictionary operations run as fast as
    a regular dictionary.

    Write to disk is delayed until close or sync (similar to gdbm's fast mode).

    Input file format is automatically discovered.
    Output file format is selectable between pickle, json, and csv.
    All three serialization formats are backed by fast C implementations.

    """

    _event_queue = Queue()
    _process_thread = None

    def __init__(self, filename, flag="c", mode=None, format="pickle", *args, **kwds):
        self.flag = flag    # r=readonly, c=create, or n=new
        self.mode = mode    # None or an octal triple like 0644
        self.format = format    # 'csv', 'json', or 'pickle'
        self.filename = filename
        if flag != "n" and os.access(filename, os.R_OK):
            fileobj = open(filename, "rb" if format == "pickle" else "r")
            with fileobj:
                self._load(fileobj)

        if PersistentDict._process_thread is None:
            PersistentDict._process_thread = Thread(target=PersistentDict._process_loop)
            PersistentDict._process_thread.daemon = (
                True    # Don't wait on thread to exit.
            )
            PersistentDict._process_thread.start()

        dict.__init__(self, *args, **kwds)

    @staticmethod
    def _process_loop():
        while True:
            (
                filename,
                contents,
                format,
                mode,
            ) = PersistentDict._event_queue.get()

            PersistentDict._update_file(filename, contents, format, mode)

    def sync(self):
        """Write dict to disk"""
        if self.flag == "r":
            return
        PersistentDict._update_file(self.filename, self, self.format, self.mode)

    def async_sync(self):
        """Write dict to disk via worker thread. Don't mix with sync if it can be helped"""
        if self.flag == "r":
            return
        PersistentDict._event_queue.put((self.filename, deepcopy(self), self.format, self.mode))

    @staticmethod
    def _update_file(filename, contents, format, mode):
        # If we are empty delete the store if it exists.
        if not contents:
            try:
                os.remove(filename)
            except OSError:
                pass
            return

        tempname = filename + ".tmp"
        fileobj = open(tempname, "wb" if format == "pickle" else "w")
        try:
            PersistentDict._dump(fileobj, contents, format)
        except Exception:
            os.remove(tempname)
            _log.error("Unable to sync to file {}".format(filename))
        finally:
            fileobj.close()
        if os.path.exists(tempname):
            shutil.move(tempname, filename)    # atomic commit

        if mode is not None:
            os.chmod(filename, mode)

    def close(self):
        self.sync()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    @staticmethod
    def _dump(fileobj, contents, format):
        if format == "csv":
            csv.writer(fileobj).writerows(contents.items())
        elif format == "json":
            jsonapi.dump(contents, fileobj, separators=(",", ":"))
        elif format == "pickle":
            pickle.dump(dict(contents), fileobj, 2)
        else:
            raise NotImplementedError("Unknown format: " + repr(PersistentDict.format))

    def _load(self, fileobj):
        # try formats from most restrictive to least restrictive
        for loader in (pickle.load, jsonapi.load, csv.reader):
            fileobj.seek(0)
            try:
                return self.update(loader(fileobj))
            except Exception:
                pass
        raise ValueError("File not in a supported format")


if __name__ == "__main__":
    import random

    # Make and use a persistent dictionary
    with PersistentDict("/tmp/demo.json", "c", format="json") as d:
        print(d, "start")
        d["abc"] = "123"
        d["rand"] = random.randrange(10000)
        print(d, "updated")

    # Show what the file looks like on disk
    with open("/tmp/demo.json", "rb") as f:
        print(f.read())



from __future__ import annotations

import importlib
import inspect
import logging
from types import ModuleType
from typing import List, Type

__all__: List[str] = ["get_module", "get_class", "get_subclasses"]

_log = logging.getLogger(__name__)


def get_module(module: str) -> ModuleType:
    """Returns a dynamically loaded module. If not found on pythonpath, then raise a ModuleNotFound error.
    This method is a wrapper around Python's builtin function, importlib.import_module(...).
    See https://docs.python.org/3/library/importlib.html#importlib.import_module

    :param module: The name argument specifies what module to import in absolute or relative terms (e.g. either pkg.mod or ..mod).
        If the name is specified in relative terms, then the package argument must be set to the name of the package which is to act as the
        anchor for resolving the package name (e.g. import_module('..mod', 'pkg.subpkg') will import pkg.mod).
    :type module: str
    :raises ModuleNotFoundError:
    :return: A module
    :rtype: ModuleType
    """
    try:
        return importlib.import_module(module)
    except ModuleNotFoundError as e:
        _log.debug(f"Module: {module} not found. Make sure it is on the PYTHONPATH")
        raise e


def get_class(module: str | ModuleType, class_name: str) -> Type:
    """Retrieve a Type from a module. If module is a string, attempt to load it via importlib.import_module.
    If not a string, then directly look for a class within the passed module.

    :param module: the path to a module or the actual module
    :type module: str | ModuleType
    :param class_name: the name of the type in the module
    :type class_name: str
    :raises AttributeError:
    :return: Returns the class from the module
    :rtype: Type
    """
    try:
        if isinstance(module, str):
            return getattr(get_module(module), class_name)
        return getattr(module, class_name)
    except AttributeError as e:
        _log.debug(f"Class {class_name} is not defined in {module}.")
        raise e


def get_subclasses(module: ModuleType | str,
                   parent_class: Type | str,
                   return_all=False) -> List[Type]:
    """Returns a list of subclasses of a specific type. If return_all is set to True,
    returns all subclasses, otherwise return a list with only the first subclass found.

    :param module: A module containing classes
    :type module: ModuleType | str
    :param parent_class: The parent class that could be a parent of classes in the module
    :type parent_class: Type | str
    :param return_all: True if all subclasses are desired; False if only the first subclass. Defaults to False
    :type return_all: bool, optional
    :raises ValueError: Raises ValueError if no subclasses are found.
    :return: A list of sublcasses of a specific type
    :rtype: List
    """
    all_subclasses = []

    if isinstance(module, str):
        module = importlib.import_module(module)
    if isinstance(parent_class, str):
        parent_class = getattr(module, parent_class)
    for c in inspect.getmembers(module, inspect.isclass):
        if parent_class in c[1].__bases__:
            all_subclasses.append(c[1])
            if not return_all:
                break

    if not all_subclasses:
        raise ValueError(f"No subclass of {parent_class} found in {module.__name__}")

    return all_subclasses



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# green
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""VIP - VOLTTRON™ Interconnect Protocol implementation

See https://volttron.readthedocs.io/en/develop/core_services/messagebus/VIP/VIP-Overview.html
for protocol specification.

This module is useful for using VIP outside of gevent. Please understand
that ZeroMQ sockets are not thread-safe and care must be used when using
across threads (or avoided all together). There is no locking around the
state as there is with the gevent version in the green sub-module.
"""

from threading import local as _local

import zmq as _zmq

from volttron.utils.socket import _Socket


class Socket(_Socket, _zmq.Socket):
    _context_class = _zmq.Context
    _local_class = _local


class BaseConnection(object):
    """
    Base connection class for message bus connection.
    """

    def __init__(self, url, identity, instance_name):
        self._url = url
        self._identity = identity
        self._instance_name = instance_name
        self._vip_handler = None



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

__all__ = [
    "execute_command", "vip_main", "is_volttron_running", "wait_for_volttron_startup",
    "wait_for_volttron_shutdown", "start_agent_thread", "isapipe"
]

import logging
import os
import subprocess
import stat
import sys

import gevent
import psutil

_log = logging.getLogger(__name__)


def execute_command(cmds, env=None, cwd=None, logger=None, err_prefix=None) -> str:
    """Executes a command as a subprocess (Not greenlet safe!)

    If the return code of the call is 0 then return stdout otherwise
    raise a RuntimeError.  If logger is specified then write the exception
    to the logger otherwise this call will remain silent.

    :param cmds:list of commands to pass to subprocess.run
    :param env: environment to run the command with
    :param cwd: working directory for the command
    :param logger: a logger to use if errors occure
    :param err_prefix: an error prefix to allow better tracing through the error message
    :return: stdout string if successful

    :raises RuntimeError: if the return code is not 0 from suprocess.run
    """

    results = subprocess.run(cmds,
                             env=env,
                             cwd=cwd,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
    if results.returncode != 0:
        err_prefix = err_prefix if err_prefix is not None else "Error executing command"
        err_message = ("\n{}: Below Command failed with non zero exit code.\n"
                       "Command:{} \nStdout:\n{}\nStderr:\n{}\n".format(
                           err_prefix, results.args, results.stdout, results.stderr))
        if logger:
            logger.exception(err_message)
            raise RuntimeError()
        else:
            raise RuntimeError(err_message)

    return results.stdout.decode("utf-8")


def start_agent_thread(cls, **kwargs):
    """Instantiate an agent class and run it in a new daemon thread.

    Returns the thread object.
    """
    import threading

    agent = cls(**kwargs)
    thread = threading.Thread(target=agent.run)
    thread.daemon = True
    thread.start()
    return thread


def isapipe(fd):
    fd = getattr(fd, "fileno", lambda: fd)()
    return stat.S_ISFIFO(os.fstat(fd).st_mode)


def vip_main(agent_class, identity=None, version="0.1", **kwargs):
    """Default main entry point implementation for VIP agents."""
    from volttron.utils import (ClientContext as cc, is_valid_identity, get_address)
    try:
        # If stdout is a pipe, re-open it line buffered
        if isapipe(sys.stdout):
            # Hold a reference to the previous file object so it doesn't
            # get garbage collected and close the underlying descriptor.
            stdout = sys.stdout
            sys.stdout = os.fdopen(stdout.fileno(), "w", 1)

        # Quiet printing of KeyboardInterrupt by greenlets
        Hub = gevent.hub.Hub
        Hub.NOT_ERROR = Hub.NOT_ERROR + (KeyboardInterrupt, )

        config = os.environ.get("AGENT_CONFIG")
        identity = os.environ.get("AGENT_VIP_IDENTITY", identity)
        publickey = kwargs.pop("publickey", None)
        if not publickey:
            publickey = os.environ.get("AGENT_PUBLICKEY")
        secretkey = kwargs.pop("secretkey", None)
        if not secretkey:
            secretkey = os.environ.get("AGENT_SECRETKEY")
        serverkey = kwargs.pop("serverkey", None)
        if not serverkey:
            serverkey = os.environ.get("VOLTTRON_SERVERKEY")

        # AGENT_PUBLICKEY and AGENT_SECRETKEY must be specified
        # for the agent to execute successfully.  aip should set these
        # if the agent is run from the platform.  If run from the
        # run command it should be set automatically from vctl and
        # added to the server.
        #
        # TODO: Make required for all agents.  Handle it through vctl and aip.
        if not os.environ.get("_LAUNCHED_BY_PLATFORM"):
            if not publickey or not secretkey:
                raise ValueError("AGENT_PUBLIC and AGENT_SECRET environmental variables must "
                                 "be set to run without the platform.")

        message_bus = os.environ.get("MESSAGEBUS", "zmq")
        if identity is not None:
            if not is_valid_identity(identity):
                _log.warning("Deprecation warining")
                _log.warning(f"All characters in {identity} are not in the valid set.")

        address = get_address()
        agent_uuid = os.environ.get("AGENT_UUID")
        volttron_home = cc.get_volttron_home()

        # TODO Bring back certs
        # from volttron.client.certs import Certs
        # certs = Certs()
        if agent_class.__name__ == "Agent":
            agent = agent_class(config_path=config,
                                identity=identity,
                                address=address,
                                agent_uuid=agent_uuid,
                                volttron_home=volttron_home,
                                version=version,
                                message_bus=message_bus,
                                publickey=publickey,
                                secretkey=secretkey,
                                serverkey=serverkey,
                                **kwargs)
        else:
            agent = agent_class(config_path=config,
                                identity=identity,
                                address=address,
                                agent_uuid=agent_uuid,
                                volttron_home=volttron_home,
                                version=version,
                                message_bus=message_bus,
                                publickey=publickey,
                                secretkey=secretkey,
                                serverkey=serverkey,
                                **kwargs)

        try:
            run = agent.run
        except AttributeError:
            run = agent.core.run
        task = gevent.spawn(run)
        try:
            task.join()
        finally:
            task.kill()
    except KeyboardInterrupt:
        pass


def is_volttron_running(volttron_home):
    """
    Checks if volttron is running for the given volttron home. Checks if a VOLTTRON_PID file exist and if it does
    check if the PID in the file corresponds to a running process. If so, returns True else returns False
    :param vhome: volttron home
    :return: True if VOLTTRON_PID file exists and points to a valid process id
    """

    pid_file = os.path.join(volttron_home, "VOLTTRON_PID")
    if os.path.exists(pid_file):
        running = False
        with open(pid_file, "r") as pf:
            pid = int(pf.read().strip())
            running = psutil.pid_exists(pid)
        return running
    else:
        return False


def wait_for_volttron_startup(vhome, timeout):
    # Check for VOLTTRON_PID
    sleep_time = 0
    while (not is_volttron_running(vhome)) and sleep_time < timeout:
        gevent.sleep(3)
        sleep_time += 3
    if sleep_time >= timeout:
        raise Exception("Platform startup failed. Please check volttron.log in {}".format(vhome))


def wait_for_volttron_shutdown(vhome, timeout):
    # Check for VOLTTRON_PID
    sleep_time = 0
    while (is_volttron_running(vhome)) and sleep_time < timeout:
        gevent.sleep(1)
        sleep_time += 1
    if sleep_time >= timeout:
        raise Exception(
            "Platform shutdown failed. Please check volttron.cfg.log in {}".format(vhome))



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
"""Run gevent Greenlets in their own threads.

Supports killing threads and executing callbacks from other threads.
"""

import functools
import sys
import threading

import gevent
from gevent import GreenletExit

__all__ = ["AsyncCall", "GreenletExit"]

__author__ = "Brandon Carpenter <brandon.carpenter@pnnl.gov>"


class AsyncCall(object):
    """Send functions to another thread's gevent hub for execution."""

    def __init__(self, hub=None):
        """Install this async handler in a hub.

        If hub is None, the current thread's hub is used.
        """
        if hub is None:
            hub = gevent.get_hub()
        self.calls = calls = []
        self.__async = hub.loop.async_()
        self.__async.start(functools.partial(self._run_calls, calls))

    def __del__(self):
        """Stop the async handler on deletion."""
        print("deleted")
        self.__async.stop()

    def send(self, receiver, func, *args, **kwargs):
        """Send a function to the hub to be called there.

        All the arguments to this method are placed in a queue and the
        hub is signaled that a function is ready. When the hub switches
        to this handler, the functions are iterated over, each being
        called with its results sent to the receiver.

        func is called with args and kwargs in the thread of the
        associated hub. If receiver is None, results are ignored and
        errors are printed when exceptions occur. Otherwise, receiver is
        called with the 2-tuple (exc_info, result). If an unhandled
        exception occurred, exc_info is the 3-tuple returned by
        sys.exc_info() and result is None. Otherwise exc_info is None
        and the result is func's return value.

        Note that receiver is called from the hub's thread and may
        need to be injected into the thread of the receiver.
        """
        self.calls.append((receiver, func, args, kwargs))
        self.__async.send()

    @staticmethod
    def _run_call(receiver, func, args, kwargs):
        """Run a pending call in its own greenlet."""
        try:
            exc_info, result = None, func(*args, **kwargs)    # pylint: disable=star-args
        except Exception:    # pylint: disable=broad-except
            exc_info, result = sys.exc_info(), None
        if receiver is not None:
            receiver((exc_info, result))
        elif exc_info:
            hub = gevent.get_hub()
            hub.handle_error(func, *exc_info)    # pylint: disable=star-args

    # This method is static to prevent a reference loop so the object
    # can be garbage collected without stopping the async handler.
    @classmethod
    def _run_calls(cls, calls):
        """Execute pending calls."""
        while calls:
            args = calls.pop()
            gevent.spawn(cls._run_call, *args)    # pylint: disable=star-args



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import logging
import os
import sys
import re
import traceback

_log = logging.getLogger(__name__)


def is_ip_private(vip_address):
    """Determines if the passed vip_address is a private ip address or not.

    :param vip_address: A valid ip address.
    :return: True if an internal ip address.
    """
    ip = vip_address.strip().lower().split("tcp://")[1]

    # https://en.wikipedia.org/wiki/Private_network

    priv_lo = re.compile(r"^127.\d{1,3}.\d{1,3}.\d{1,3}$")
    priv_24 = re.compile(r"^10.\d{1,3}.\d{1,3}.\d{1,3}$")
    priv_20 = re.compile(r"^192.168.\d{1,3}.\d{1,3}$")
    priv_16 = re.compile(r"^172.(1[6-9]|2[0-9]|3[0-1]).[0-9]{1,3}.[0-9]{1,3}$")

    return (priv_lo.match(ip) is not None or priv_24.match(ip) is not None
            or priv_20.match(ip) is not None or priv_16.match(ip) is not None)


def get_hostname():
    with open("/etc/hostname") as fp:
        hostname = fp.read().strip()

    if not hostname:
        raise ValueError("/etc/hostname file not found!")
    return hostname


def get_address(verify_listening=False):
    """Return the VIP address of the platform
    If the VOLTTRON_VIP_ADDR environment variable is set, it is used to connect to.
    Otherwise, it is derived from get_home()."""
    from volttron.utils import ClientContext as cc

    address = os.environ.get("VOLTTRON_VIP_ADDR")
    if not address:
        # Connect via virtual unix socket if linux platform (mac doesn't have @ in it)
        abstract = "@" if sys.platform.startswith("linux") else ""
        address = "ipc://%s%s/run/vip.socket" % (
            abstract,
            cc.get_volttron_home(),
        )

    import zmq.green as zmqgreen
    import zmq

    # The following block checks to make sure that we can
    # connect to the zmq based upon the ipc address.
    #
    # The zmq.sock.bind() will raise an error because the
    # address is already bound (therefore volttron is running there)
    sock = None
    try:
        # TODO: We should not just do the connection test when verfiy_listening is True but always
        # Though we leave this here because we have backward compatible unit tests that require
        # the get_address to not have somethiing bound to the address.
        if verify_listening:
            ctx = zmqgreen.Context.instance()
            sock = ctx.socket(zmq.PUB)    # or SUB - does not make any difference
            sock.bind(address)
            raise ValueError("Unable to connect to vip address "
                             f"make sure VOLTTRON_HOME: {cc.get_volttron_home()} "
                             "is set properly")
    except zmq.error.ZMQError as e:
        _log.error(f"Zmq error was {e}\n{traceback.format_exc()}")
    finally:
        try:
            sock.close()
        except AttributeError as e:    # Raised when sock is None type
            pass

    return address



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import logging
import os
from typing import Union

from urllib.parse import urlparse
import weakref

from .base import SubsystemBase

from volttron.client.known_identities import AUTH
from volttron.utils import ClientContext as cc
from volttron.utils import jsonapi
from volttron.utils.jsonrpc import RemoteError
from volttron.utils.keystore import KeyStore
"""
The auth subsystem allows an agent to quickly query authorization state
(e.g., which capabilities each user has been granted).
"""

__docformat__ = "reStructuredText"
__version__ = "1.1"

_log = logging.getLogger(__name__)


class Auth(SubsystemBase):

    def __init__(self, owner, core, rpc):
        self._owner = owner
        self._core = weakref.ref(core)
        self._rpc = weakref.ref(rpc)
        self._user_to_capabilities = {}
        self._dirty = True
        self._csr_certs = dict()
        self.remote_certs_dir = None

        def onsetup(sender, **kwargs):
            rpc.export(self._update_capabilities, "auth.update")

        core.onsetup.connect(onsetup, self)

    def connect_remote_platform(
        self,
        address,
        serverkey=None,
        agent_class: "Agent" = None,
        keystore: KeyStore = None,
    ) -> Union["Agent", None]:
        """
        Agent atempts to connect to a remote platform to exchange data.

        address must start with http, https, tcp, ampq, or ampqs or a ValueError will be
        raised

        If this function is successful it will return an instance of the `agent_class`
        parameter if not then this function will return None.

        If the address parameter begins with http or https
        TODO: use the known host functionality here
        the agent will attempt to use Discovery to find the values associated with it.

        Discovery should return either an rmq-address or a vip-address or both.  In
        that situation the connection will be made using zmq.  In the event that
        fails then rmq will be tried.  If both fail then None is returned from this
        function.

        """
        from volttron.client.vip.agent.utils import build_agent
        from volttron.client.vip.agent import Agent

        if agent_class is None:
            agent_class = Agent

        parsed_address = urlparse(address)
        _log.debug("Begining auth.connect_remote_platform: {}".format(address))

        value = None
        if parsed_address.scheme == "tcp":
            # ZMQ connection
            hosts = KnownHostsStore()
            temp_serverkey = hosts.serverkey(address)
            if not temp_serverkey:
                _log.info("Destination serverkey not found in known hosts file, using config")
                destination_serverkey = serverkey
            elif not serverkey:
                destination_serverkey = temp_serverkey
            else:
                if temp_serverkey != serverkey:
                    raise ValueError("server_key passed and known hosts serverkey do not match!")
                destination_serverkey = serverkey

            publickey, secretkey = self._core().publickey, self._core().secretkey
            _log.debug("Connecting using: {}".format(cc.get_fq_identity(self._core().identity)))

            value = build_agent(
                agent_class=agent_class,
                identity=cc.get_fq_identity(self._core().identity),
                serverkey=destination_serverkey,
                publickey=publickey,
                secretkey=secretkey,
                message_bus="zmq",
                address=address,
            )
        elif parsed_address.scheme in ("https", "http"):
            from volttron.client import DiscoveryInfo
            from volttron.client import DiscoveryError

            try:
                # TODO: Use known host instead of looking up for discovery info if possible.

                # We need to discover which type of bus is at the other end.
                info = DiscoveryInfo.request_discovery_info(address)
                remote_identity = "{}.{}.{}".format(
                    info.instance_name,
                    cc.get_instance_name(),
                    self._core().identity,
                )
                # if the current message bus is zmq then we need
                # to connect a zmq on the remote, whether that be the
                # rmq router or proxy.  Also note that we are using the fully qualified
                # version of the identity because there will be conflicts if
                # volttron central has more than one platform.agent connecting
                if cc.get_messagebus() == "zmq":
                    if not info.vip_address or not info.serverkey:
                        err = "Discovery from {} did not return serverkey and/or vip_address".format(
                            address)
                        raise ValueError(err)

                    _log.debug("Connecting using: {}".format(
                        cc.get_fq_identity(self._core().identity)))

                    # use fully qualified identity
                    value = build_agent(
                        identity=cc.get_fq_identity(self._core().identity),
                        address=info.vip_address,
                        serverkey=info.serverkey,
                        secretkey=self._core().secretkey,
                        publickey=self._core().publickey,
                        agent_class=agent_class,
                    )

                else:    # we are on rmq messagebus

                    # This is if both remote and local are rmq message buses.
                    if info.messagebus_type == "rmq":
                        _log.debug("Both remote and local are rmq messagebus.")
                        fqid_local = cc.get_fq_identity(self._core().identity)

                        # Check if we already have the cert, if so use it instead of requesting cert again
                        remote_certs_dir = self.get_remote_certs_dir()
                        remote_cert_name = "{}.{}".format(info.instance_name, fqid_local)
                        certfile = os.path.join(remote_certs_dir, remote_cert_name + ".crt")
                        if os.path.exists(certfile):
                            response = certfile
                        else:
                            response = self.request_cert(address, fqid_local, info)

                        if response is None:
                            _log.error("there was no response from the server")
                            value = None
                        elif isinstance(response, tuple):
                            if response[0] == "PENDING":
                                _log.info("Waiting for administrator to accept a CSR request.")
                            value = None
                        # elif isinstance(response, dict):
                        #     response
                        elif os.path.exists(response):
                            # info = DiscoveryInfo.request_discovery_info(address)
                            # From the remote platforms perspective the remote user name is
                            #   remoteinstance.localinstance.identity, this is what we must
                            #   pass to the build_remote_connection_params for a successful

                            remote_rmq_user = cc.get_fq_identity(fqid_local, info.instance_name)
                            _log.debug("REMOTE RMQ USER IS: {}".format(remote_rmq_user))
                            remote_rmq_address = (
                                self._core().rmq_mgmt.build_remote_connection_param(
                                    remote_rmq_user,
                                    info.rmq_address,
                                    ssl_auth=True,
                                    cert_dir=self.get_remote_certs_dir(),
                                ))

                            value = build_agent(
                                identity=fqid_local,
                                address=remote_rmq_address,
                                instance_name=info.instance_name,
                                publickey=self._core().publickey,
                                secretkey=self._core().secretkey,
                                message_bus="rmq",
                                enable_store=False,
                                agent_class=agent_class,
                            )
                        else:
                            raise ValueError("Unknown path through discovery process!")

                    else:
                        # TODO: cache the connection so we don't always have to ping
                        #       the server to connect.

                        # This branch happens when the message bus is not the same note
                        # this writes to the agent-data directory of this agent if the agent
                        # is installed.
                        if cc.get_messagebus() == "rmq":
                            if not os.path.exists("keystore.json"):
                                with open("keystore.json", "w") as fp:
                                    fp.write(jsonapi.dumps(KeyStore.generate_keypair_dict()))

                            with open("keystore.json") as fp:
                                keypair = jsonapi.loads(fp.read())

                        value = build_agent(
                            agent_class=agent_class,
                            identity=remote_identity,
                            serverkey=info.serverkey,
                            publickey=keypair.get("publickey"),
                            secretkey=keypair.get("secretekey"),
                            message_bus="zmq",
                            address=info.vip_address,
                        )
            except DiscoveryError:
                _log.error(
                    "Couldn't connect to {} or incorrect response returned response was {}".format(
                        address, value))

        else:
            raise ValueError(
                "Invalid configuration found the address: {} has an invalid scheme".format(
                    address))

        return value

    # TODO Figure out certs
    # def request_cert(self, csr_server, fully_qualified_local_identity, discovery_info):
    #     """ Get a signed csr from the csr_server endpoint
    #
    #     This method will create a csr request that is going to be sent to the
    #     signing server.
    #
    #     :param csr_server: the http(s) location of the server to connect to.
    #     :return:
    #     """
    #
    #     if cc.get_messagebus() != 'rmq':
    #         raise ValueError("Only can create csr for rabbitmq based platform in ssl mode.")
    #
    #     # from volttron.client.web import DiscoveryInfo
    #     config = RMQConfig()
    #
    #     if not config.is_ssl:
    #         raise ValueError("Only can create csr for rabbitmq based platform in ssl mode.")
    #
    #     # info = discovery_info
    #     # if info is None:
    #     #     info = DiscoveryInfo.request_discovery_info(csr_server)
    #
    #     certs = Certs()
    #     csr_request = certs.create_csr(fully_qualified_local_identity, discovery_info.instance_name)
    #     # The csr request requires the fully qualified identity that is
    #     # going to be connected to the external instance.
    #     #
    #     # The remote instance id is the instance name of the remote platform
    #     # concatenated with the identity of the local fully quallified identity.
    #     remote_cert_name = "{}.{}".format(discovery_info.instance_name, fully_qualified_local_identity)
    #     remote_ca_name = discovery_info.instance_name + "_ca"
    #
    #     # if certs.cert_exists(remote_cert_name, True):
    #     #     return certs.cert(remote_cert_name, True)
    #
    #     json_request = dict(
    #         csr=csr_request.decode("utf-8"),
    #         identity=remote_cert_name,  # cc.get_instance_name()+"."+self._core().identity,
    #         hostname=config.hostname
    #     )
    #     request = grequests.post(csr_server + "/csr/request_new",
    #                              json=jsonapi.dumps(json_request),
    #                              verify=False)
    #     response = grequests.map([request])
    #
    #     if response and isinstance(response, list):
    #         response[0].raise_for_status()
    #     response = response[0]
    #     # response = requests.post(csr_server + "/csr/request_new",
    #     #                          json=jsonapi.dumps(json_request),
    #     #                          verify=False)
    #
    #     _log.debug("The response: {}".format(response))
    #
    #     j = response.json()
    #     status = j.get('status')
    #     cert = j.get('cert')
    #     message = j.get('message', '')
    #     remote_certs_dir = self.get_remote_certs_dir()
    #     if status == 'SUCCESSFUL' or status == 'APPROVED':
    #         certs.save_agent_remote_info(remote_certs_dir,
    #                                      fully_qualified_local_identity,
    #                                      remote_cert_name, cert.encode("utf-8"),
    #                                      remote_ca_name,
    #                                      discovery_info.rmq_ca_cert.encode("utf-8"))
    #         os.environ['REQUESTS_CA_BUNDLE'] = os.path.join(remote_certs_dir, "requests_ca_bundle")
    #         _log.debug("Set os.environ requests ca bundle to {}".format(os.environ['REQUESTS_CA_BUNDLE']))
    #     elif status == 'PENDING':
    #         _log.debug("Pending CSR request for {}".format(remote_cert_name))
    #     elif status == 'DENIED':
    #         _log.error("Denied from remote machine.  Shutting down agent.")
    #         status = Status.build(BAD_STATUS,
    #                               context="Administrator denied remote connection.  Shutting down")
    #         self._owner.vip.health.set_status(status.status, status.context)
    #         self._owner.vip.health.send_alert(self._core().identity+"_DENIED", status)
    #         self._core().stop()
    #         return None
    #     elif status == 'ERROR':
    #         err = "Error retrieving certificate from {}\n".format(
    #             config.hostname)
    #         err += "{}".format(message)
    #         raise ValueError(err)
    #     else:  # No resposne
    #         return None
    #
    #     certfile = os.path.join(remote_certs_dir, remote_cert_name + ".crt")
    #     if os.path.exists(certfile):
    #         return certfile
    #     else:
    #         return status, message

    # TODO Figure out certs
    # def get_remote_certs_dir(self):
    #     if not self.remote_certs_dir:
    #         install_dir = os.path.join(get_home(), "agents", self._core().agent_uuid)
    #         files = os.listdir(install_dir)
    #         for f in files:
    #             agent_dir = os.path.join(install_dir, f)
    #             if os.path.isdir(agent_dir):
    #                 break  # found
    #         sub_dirs = os.listdir(agent_dir)
    #         for d in sub_dirs:
    #             d_path = os.path.join(agent_dir, d)
    #             if os.path.isdir(d_path) and d.endswith("agent-data"):
    #                 self.remote_certs_dir = d_path
    #     return self.remote_certs_dir

    def _fetch_capabilities(self):
        while self._dirty:
            self._dirty = False
            try:
                self._user_to_capabilities = (self._rpc().call(
                    AUTH, "get_user_to_capabilities").get(timeout=10))
                _log.debug("self. user to cap {}".format(self._user_to_capabilities))
            except RemoteError:
                self._dirty = True

    def get_capabilities(self, user_id):
        """Gets capabilities for a given user.

        :param user_id: user id field from VOLTTRON Interconnect Protocol
        :type user_id: str
        :returns: list of capabilities
        :rtype: list
        """
        self._fetch_capabilities()
        return self._user_to_capabilities.get(user_id, [])

    def _update_capabilities(self, user_to_capabilities):
        identity = self._rpc().context.vip_message.peer
        if identity == AUTH:
            self._user_to_capabilities = user_to_capabilities
            self._dirty = True



from configparser import ConfigParser

# used to make sure that volttron_home hasn't be modified
# since written to disk.
import logging
import os
from pathlib import Path
from typing import Optional

from .frozendict import FrozenDict

_log = logging.getLogger(__name__)


class ClientContext:
    """
    The `ClientContext` class is the single source of truth within
    a process running this system.
    """

    __volttron_home__: Optional[Path] = None
    __config__: dict = {}
    __config_keys__ = ("vip-address", "bind-web-address", "instance-name", "message-bus",
                       "web-ssl-cert", "web-ssl-key", "web-secret-key", "secure-agent-users")

    @classmethod
    def __load_config__(cls: "ClientContext"):
        if not cls.__config__:
            cls.__config__ = FrozenDict()

            volttron_home = ClientContext.get_volttron_home()
            config_file = os.path.join(volttron_home, "config")
            if os.path.exists(config_file):
                parser = ConfigParser()
                parser.read(config_file)
                options = parser.options("volttron")
                for option in options:
                    cls.__config__[option] = parser.get("volttron", option)
                cls.__config__.freeze()
        return cls.__config__

    @classmethod
    def get_config_param(cls, key: str, default: Optional[str] = None) -> Optional[str]:

        ClientContext.__load_config__()
        return cls.__config__.get(key, default)

    @classmethod
    def is_rabbitmq_available(cls):
        rabbitmq_available = True
        try:
            import pika

            rabbitmq_available = True
        except ImportError:
            os.environ["RABBITMQ_NOT_AVAILABLE"] = "True"
            rabbitmq_available = False
        return rabbitmq_available

    @classmethod
    def get_volttron_home(cls) -> str:
        """
        Return the VOLTTRON_HOME directory specified or default directory.

        If the VOLTTRON_HOME environment variable is set, it used.
        Otherwise, the default value of '~/.volttron' is used.

        If the volttron_home does not exist then this function will create
        it if possible.  This function also creates a check file for the
        VOLTTRON_HOME such that if the VOLTTRON_HOME is modified during
        runtime it will be detected and cause an error.

        @return:str:
            The absolute path to the volttron_home.
        """

        # vhome to test against for modification.
        vhome = (Path(os.environ.get("VOLTTRON_HOME", "~/.volttron")).expanduser().resolve())

        # cls variable is set the first time through this function
        # so we test to make sure nothing has changed from vhome and
        # the cls.__volttron_home__ variable.
        if cls.__volttron_home__:
            if vhome != cls.__volttron_home__:
                raise ValueError("VOLTTRON_HOME has been changed.  Possible nefarious act!")

        # Initialize class variable here and write a file inside the
        # volttron_home that we can check against.
        if cls.__volttron_home__ is None:
            cls.__volttron_home__ = vhome

            if not vhome.exists():
                # python 3.6 doesn't support pathlike object in mkdir
                os.makedirs(str(vhome), exist_ok=True)

        return str(vhome)

    @classmethod
    def get_fq_identity(cls, identity, platform_instance_name=None):
        """
        Return the fully qualified identity for the passed core identity.

        Fully qualified identities are instance_name.identity

        :param identity:
        :param platform_instance_name: str The name of the platform.
        :return:
        """
        if not platform_instance_name:
            platform_instance_name = cls.get_config_param("instance-name")
        return f"{platform_instance_name}.{identity}"

    @classmethod
    def get_messagebus(cls):
        """Get type of message bus - zeromq or rabbbitmq."""
        return cls.get_config_param("message-bus")

    @classmethod
    def get_instance_name(cls):
        """Get type of message bus - zeromq or rabbbitmq."""
        instance_name = cls.get_config_param('instance-name', None)
        if instance_name is not None:
            instance_name = instance_name.strip('"')

        if not instance_name:
            _log.warning("Using hostname as instance name.")
            if os.path.isfile('/etc/hostname'):
                with open('/etc/hostname') as f:
                    instance_name = f.read().strip()

        return instance_name

    @classmethod
    def is_web_enabled(cls):
        """Returns True if web enabled, False otherwise"""
        if cls.get_config_param("bind-web-address"):
            return True
        return False

    @classmethod
    def is_secure_mode(cls):
        """Returns True if running in secure mode, False otherwise"""
        secure_mode = cls.get_config_param("secure-agent-users", False)
        if secure_mode:
            secure_mode = secure_mode.upper() == "TRUE"
        return secure_mode

    @classmethod
    def get_server_key(cls):
        """Returns server key"""
        from volttron.utils.keystore import KeyStore
        keystore_path = os.path.join(cls.get_volttron_home(), "keystore")
        keystore = KeyStore(keystore_path)
        return keystore.public

    @classmethod
    def get_agent_keys(cls, vip_id):
        from volttron.utils.keystore import KeyStore
        keystore_path = os.path.join(cls.get_volttron_home(), "agents", vip_id, "keystore.json")
        keystore = KeyStore(keystore_path)
        return keystore.public, keystore.secret



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:

# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
# }}}
"""Utilities for tracking VIP message statistics at the router."""

import gevent

from volttron.server.router import UNROUTABLE, ERROR, INCOMING

__all__ = ["Tracker"]


def pick(frames, index):
    """Return the frame at index, converted to bytes, or None."""
    try:
        return bytes(frames[index])
    except IndexError:
        return None


def increment(prop, key):
    """Increment or set to 1 the value in prop[key]."""
    try:
        prop[key] += 1
    except KeyError:
        prop[key] = 1


class Tracker(object):
    """Object for sharing data between the router and control objects."""

    def __init__(self):
        self._reset()
        self.enabled = False

    def reset(self):
        """Reset all counters to default values and set start time."""
        self._reset()
        self.stats["start"] = gevent.get_hub().loop.now()

    def _reset(self):
        """Initialize statistics counters."""
        self.stats = {
            "error": {
                "error": {},
                "peer": {},
                "user": {},
                "subsystem": {}
            },
            "unroutable": {
                "error": {},
                "peer": {}
            },
            "incoming": {
                "peer": {},
                "user": {},
                "subsystem": {}
            },
            "outgoing": {
                "peer": {},
                "user": {},
                "subsystem": {}
            },
        }

    def hit(self, topic, frames, extra):
        """Increment counters for given topic and frames."""
        if self.enabled:
            if topic == UNROUTABLE:
                stat = self.stats["unroutable"]
                increment(stat["error"], extra)
            else:
                user = pick(frames, 3)
                subsystem = pick(frames, 5)
                if topic == ERROR:
                    stat = self.stats["error"]
                    increment(stat["error"], bytes(extra[0]))
                else:
                    stat = self.stats["incoming" if topic == INCOMING else "outgoing"]
                increment(stat["user"], user)
                increment(stat["subsystem"], subsystem)
            increment(stat["peer"], pick(frames, 0))

    def enable(self):
        """Enable tracking."""
        if not self.enabled:
            self.reset()
            self.enabled = True

    def disable(self):
        """Disable tracking."""
        if self.enabled:
            self.enabled = False
            self.stats["end"] = gevent.get_hub().loop.now()



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

__all__ = ["create_file_if_missing"]

import os
import errno
import logging

_log = logging.getLogger(__name__)


def create_file_if_missing(path, permission=0o660, contents=None):
    dirname = os.path.dirname(path)
    if dirname and not os.path.exists(dirname):
        try:
            os.makedirs(dirname)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    try:
        with open(path) as fd:
            pass
    except IOError as exc:
        if exc.errno != errno.ENOENT:
            raise
        _log.debug("missing file %s", path)
        _log.info("creating file %s", path)
        fd = os.open(path, os.O_CREAT | os.O_WRONLY, permission)
        success = False
        try:
            if contents:
                contents = (contents if isinstance(contents, bytes) else contents.encode("utf-8"))
                os.write(fd, contents)
                success = True
        except Exception as e:
            raise e
        finally:
            os.close(fd)
        return success



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:

# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
# }}}
"""Module for storing local public and secret keys and remote public keys"""
import base64
import binascii
import logging
import os
import urllib

from zmq.utils import z85
from zmq import curve_keypair

from volttron.utils import jsonapi, ClientContext as cc
from volttron.utils import create_file_if_missing

_log = logging.getLogger(__name__)

BASE64_ENCODED_CURVE_KEY_LEN = 43


def get_server_keys():
    try:
        # attempt to read server's keys. Should be used only by multiplatform connection and tests
        # If agents such as forwarder attempt this in secure mode this will throw access violation exception
        ks = KeyStore()
    except IOError as e:
        raise RuntimeError(
            "Exception accessing server keystore. Agents must use agent's public and private key"
            "to build dynamic agents when running in secure mode. Exception:{}".format(e))

    return ks.public, ks.secret


def get_random_key(length: int = 65) -> str:
    """
    Returns a hex random key of specified length.  The length must be > 0 in order for
    the key to be valid.  Raises a ValueError if the length is invalid.

    The default length is 65, which is 130 in length when hexlify is run.

    :param length:
    :return:
    """
    if length <= 0:
        raise ValueError("Invalid length specified for random key must be > 0")

    import binascii

    random_key = binascii.hexlify(os.urandom(length)).decode("utf-8")
    return random_key


def encode_key(key):
    """Base64-encode and return a key in a URL-safe manner."""
    # There is no easy way to test if key is already base64 encoded and ASCII decoded. This seems the best way.
    if len(key) % 4 != 0:
        return key
    key = key if isinstance(key, bytes) else key.encode("utf-8")
    try:
        assert len(key) in (32, 40)
    except AssertionError:
        raise AssertionError("Assertion error while encoding key:{}, len:{}".format(key, len(key)))
    if len(key) == 40:
        key = z85.decode(key)
    return base64.urlsafe_b64encode(key)[:-1].decode("ASCII")


def decode_key(key):
    """Parse and return a Z85 encoded key from other encodings."""
    if isinstance(key, str):
        key = key.encode("ASCII")
    length = len(key)
    if length == 40:
        return key
    elif length == 43:
        return z85.encode(base64.urlsafe_b64decode(key + "=".encode("ASCII")))
    elif length == 44:
        return z85.encode(base64.urlsafe_b64decode(key))
    elif length == 54:
        return base64.urlsafe_b64decode(key + "==".encode("ASCII"))
    elif length == 56:
        return base64.urlsafe_b64decode(key)
    elif length == 64:
        return z85.encode(binascii.unhexlify(key))
    elif length == 80:
        return binascii.unhexlify(key)
    raise ValueError("unknown key encoding")


class BaseJSONStore(object):
    """JSON-file-backed store for dictionaries"""

    def __init__(self, filename, permissions=0o600):
        self.filename = filename
        self.permissions = permissions
        try:
            created = create_file_if_missing(filename, contents="{}")
            if created:
                # remove access to group
                os.chmod(filename, permissions)
        except Exception as e:
            import traceback

            _log.error(traceback.print_exc())
            raise RuntimeError("Failed to access KeyStore: {}".format(filename))

    def store(self, data):
        fd = os.open(
            self.filename,
            os.O_CREAT | os.O_WRONLY | os.O_TRUNC,
            self.permissions,
        )
        try:
            os.write(fd, jsonapi.dumpb(data, indent=4))
        finally:
            os.close(fd)

    def load(self):
        try:
            with open(self.filename, "r") as json_file:
                return jsonapi.load(json_file)
        except ValueError:
            # If the file is empty jsonapi.load will raise ValueError
            return {}

    def remove(self, key):
        data = self.load()
        try:
            del data[key]
        except KeyError as e:
            msg = 'Key "{}" is not present in {}'.format(key, self.filename)
            raise KeyError(msg)
        else:
            self.store(data)

    def update(self, new_data):
        data = self.load()
        data.update(new_data)
        self.store(data)


class KeyStore(BaseJSONStore):
    """Handle generation, storage, and retrival of CURVE key pairs"""

    def __init__(self, filename=None, encoded_public=None, encoded_secret=None):
        if filename is None:
            filename = self.get_default_path()
        super(KeyStore, self).__init__(filename)
        if not self.isvalid():
            if encoded_public and encoded_secret:
                self.store({
                    "public": encoded_public,
                    "secret": encode_key(encoded_secret),
                })
            else:
                _log.debug("calling generate from keystore")
                self.generate()

    @staticmethod
    def get_default_path():
        return os.path.join(cc.get_volttron_home(), "keystore")

    @staticmethod
    def get_agent_keystore_path(identity=None):
        if identity is None:
            raise AttributeError("invalid identity")
        return os.path.join(cc.get_volttron_home(), f"keystores/{identity}/keystore.json")

    @staticmethod
    def generate_keypair_dict():
        """Generate and return new keypair as dictionary"""
        public, secret = curve_keypair()
        encoded_public = encode_key(public)
        encoded_secret = encode_key(secret)
        attempts = 0
        max_attempts = 3

        done = False
        while not done and attempts < max_attempts:
            # Keys that start with '-' are hard to use and cause issues with the platform
            if encoded_secret.startswith("-") or encoded_public.startswith("-"):
                # try generating public and secret key again
                public, secret = curve_keypair()
                encoded_public = encode_key(public)
                encoded_secret = encode_key(secret)
            else:
                done = True

        return {"public": encoded_public, "secret": encoded_secret}

    def generate(self):
        """Generate and store new key pair"""
        self.store(self.generate_keypair_dict())

    def _get_key(self, keyname):
        """Get key and make sure it's type is str (not unicode)

        The json module returns all strings as unicode type, but base64
        decode expects byte type as input. The conversion from unicode
        type to str type is safe in this case, because encode_key
        returns str type (ASCII characters only).
        """
        key = self.load().get(keyname, None)
        if key:
            try:
                key.encode("ascii")
            except UnicodeEncodeError:
                _log.warning("Non-ASCII character found for key {} in {}".format(
                    keyname, self.filename))
                key = None
        return key

    @property
    def public(self):
        """Return encoded public key"""
        return self._get_key("public")

    @property
    def secret(self):
        """Return encoded secret key"""
        return self._get_key("secret")

    def isvalid(self):
        """Check if key pair is valid"""
        return self.public and self.secret


class KnownHostsStore(BaseJSONStore):
    """Handle storage and retrival of known hosts"""

    def __init__(self, filename=None):
        if filename is None:
            filename = f"{cc.get_volttron_home()}/known_hosts"
        # all agents need read access to known_hosts file
        super(KnownHostsStore, self).__init__(filename, permissions=0o644)

    def add(self, addr, server_key):
        self.update({self._parse_addr(addr): server_key})

    def serverkey(self, addr):
        return self.load().get(self._parse_addr(addr), None)

    @staticmethod
    def _parse_addr(addr):
        url = urllib.parse.urlparse(addr)
        if url.netloc:
            return url.netloc
        return url.path
